
(* ======================================================================
    DYNAMICKE SEZNAMY - FRONTA
   ===================================================================
*)

program Dynamicka_fronta;
uses crt;

const nula = 0;           (* takto nadeklarovane typy a constanty a umozni *)
type typklice = byte;           (* pozdejsi prehledne deklarace a pripadne *)
                                               (* zmeny pri ruznem pouziti *)
type dynprm = ^prvek;
      prvek = record klic : typklice;
                     nasl : dynprm;
              end;

type zasobnik = record vrch : dynprm;       (* schvalne takhle slozite pro *)
                end;                        (* zrejmost analogie           *)
type fronta   = record celo, konec : dynprm;
                end;

(* v programu je pak mozne pouzit napr promenne techto typu :              *)

var dz : zasobnik;                         (* jiste by slo var dz : dynprm *)
    df : fronta;                           (* a df_celo, df_konec : dynprm *)
   pom : dynprm;   (* takto nasledne prehlednejsi program, hlavne u fronty *)
                                                   (* a momentalne pro nas *)

{
   =========================================================================
   tvorba zasobniku - opakovani - posloupnost prikazu:
   inicializace : vrch:=nil;

   nasledne prikazy z procedury ulozeni prvku na vrchol zasobniku :
   new(pom); pom^.klic:=x; pom^.nasl:=vrch; vrch:=pom;

             1      2      3             new(pom) => .klic:=x
=>  nil <--- * <--- * <--- * <--- vrch               .nasl:=vrch  <--- vrch
-----------------------------------------------------------------------------
   tvorba fronty probiha podobne, rozsirovani na konci
   => nutno pouzit dva ukazatele
      mejme obecne uz nejak naplnenou frontu napriklad tohoto tvaru :

                1     2     3     4
   df.celo ---> * --> * --> * --> * <--- df.konec

   posloupnost prikazu pro rozsireni fronty by mohla vypadat takto :

   new(df.konec^.nasl); df.konec:=df.konec^.nasl; df.konec^.klic:=x;  - (1)

   doposud vse v poradku
   ---------------------
   nutno vsak dat pozor na rozsirovani fronty obdobnou posloupnosti prikazu
   jako u zasobniku vzhledem k pouzite inicializaci
   !!! nemusi dojit k zretezeni !!!!
   => nutno rozlisit dva pripady zavisle na druhu inicializace.

   kde je problem ? - predevsim - jak bude vypadat inicializace ?

   procedure InitF(var df : fronta);
   begin df.celo:=nil; df.konec:=df.celo;                             - (2)
   end;

   Pokud pouziji tuto inicializaci a nasledne posloupnost prikazu (1),
   co se stane pri pridani prvniho prvku a nasledne v cyklu prvku dalsich ?

   ==========================================================================

   nutny jiny postup - dve varianty zavisle na tvaru inicializace :
1. varianta :
   jina inicializace - alokace 'nuloveho' prvku, ktery bude vzdy koncovy a ktery
                       lze pripadne vyuzit pri prohledavani pomoci zarazky

   procedure InitF(var df : fronta);                                  - (3)
   begin with df do
         begin new(celo); celo^.klic:=nula; celo^.nasl:=nil;
               konec:=celo;
         end;
   end;

   nulovani koncoveho prvku je vhodnejsi pro pozdejsi opakovani prenechat
   zvlastni procedure, takze napriklad takto :

   procedure Nuluj(var ukaz : dynprm);
   begin ukaz^.klic:=nula; ukaz^.nasl:=nil;
   end;
   procedure InitF(var df : fronta);
   begin with df do begin new(celo); Nuluj(celo); konec:=celo; end;
   end;
   ------------------------------------------------------------------------
   situace tedy vypada takto :

             0
   celo ---> * <--- konec        =>        posloupnost prikazu na rozsireni :

   konec^.klic:=x; new(konec^.nasl); konec:=konec^.nasl;              - (4)
   Nuluj(konec);

             1     2     3     4     0
   celo ---> * --> * --> * --> * --> * -- > nil
                                     |
                              konec /

   volny-prazdny prvek na konci lze vyuzit jako zarazka pri vyhledavani
   => zjednoduseni podminky u prikazu cyklu pro vyhledani prvku - viz nize;

   pokud by mi vadila 0 na konci - pak :

2. varianta :
   nutno rozlisit, zda pridavam do prazdne fronty prvni prvek (napr.po InitF)
   nebo pozdeji za jiz nejak naplnenou frontu
=> posloupnost prikazu na rozsirovani fronty musi mit rozlisovaci podminku
   - opet dve moznosti :

   2.1 - na zaklade inicializace (3) - vytvoreni nuloveho prvku :
   with df do
   begin if celo^.klic=0
            then celo^.klic:=x
            else begin new(konec^.nasl); konec:=konec^.nasl; konec^.klic:=x;
                 end;
   end;

   2.2 - na zaklade inicializace (2) - nastaveni df.cela na nil :
   with df do
   begin if celo=nil
            then begin new(celo); celo^.klic:=x; konec:=celo; end;
            else begin new(konec^.nasl); konec:=konec^.nasl; konec^.klic:=x;
                 end;
   end;

   V obou pripadech (2.1 i 2.2) vsak pri vlozeni prvku VZDY testuji podminku
   pri castem vkladani prvku zbytecne zdrzeni
   =>

   BUDEME PREFEROVAT inicializaci (3) a rozsirovani fronty pomoci posl. (4)
   - jednodussi + vyuziti nuloveho prvku jako zarazky
  ========================================================================

  ZARAZKA - vyuziti
  motivace - nastaveni ukazatele na prvek s danym klicem - nalezeni prvku

  prirozene reseni - pro frontu s nulovym prvkem na konci :

  procedure Najdi(prvek : typk; df : fronta; var ukaz : dynprm);
  begin ukaz:=df.celo;
      while (ukaz^.klic<>prvek) and (ukaz<>df.konec) do ukaz:=ukaz^.nasl;
      if ukaz=df.konec then write(' nenalezen ')
                       else write(' hledany prvek je : ',ukaz^.klic);
  end;

  - nevyhoda - slozena podminka u cyklu while - pro dlouhe seznamy-fronty
               zvysena slozitost
  =>  reseni : vyuziti posledniho nuloveho prvku jako zarazky :

  procedure Najdi(prvek : typk; df : fronta; var ukaz : dynprm);    (* hleda *)
  begin ukaz:=df.celo; df.konec^.klic:=prvek;    (* prvek s vyuzitim zarazky *)
      while ukaz^.klic<>prvek do ukaz:=ukaz^.nasl;
      if ukaz=df.konec then write(' nenalezen ')
                       else write(' hledany prvek je : ',ukaz^.klic);
      df.konec^.klic:=nula;
  end;
  ========================================================================
}

(* nasleduje nabidka procedur a funkci pro praci s frontou
   s kratkym ukazkovym programem
*)
procedure InitF(var df : fronta);
begin new(df.celo); df.konec:=df.celo;      (* => nastavi celo^.klic:=0;   *)
end;                                        (*            celo^.nasl:=nil; *)

procedure PridejKonec(prvek : typklice; var df : fronta);
begin with df do begin konec^.klic:=prvek; new(konec^.nasl);
                       konec:=konec^.nasl;
                 end;
end;

procedure OdeberCelo(var prvek : typklice; var df : fronta);
begin  (* dodelat samostatne
	  do promenne prvek nacte hodnotu klice z cela fronty;
          ukazatel cela fronty posune o jeden prvek seznamu ;
          uvolni alokovane misto odebrane dyn.promenne
          - ukol :
            porovnat s posloupnosti prikazu pro odebrani vrcholu zasobniku
       *)
end;

function PrazdnaF(df : fronta) : boolean;
begin (* test zda ukazatel cela = ukazatel konce *)
      (* opet nutno vztahnout k inicializaci a ptat se,
         zda chci mit prazdnou frontu v pripade odebrani vech vlozenych
         promennych se zachovanim volneho koncoveho prvku pro pripadne
         dalsi pridavani nebo zda prazdna fronta bude znamenat uplne
         uvolneni vsech alokovanych promennych
         - pri pripadnem opetovnem pridavani v ramci jednoho behu programu
           - algoritmu - nutno provest novou inicializaci
      *)
end;


var x : typklice;
    z : char;

begin textattr:=58; clrscr; Randomize;
(* maly test na praci s frontou *)
  repeat
      InitF(df);                                                (* dle (3) *)
      repeat x:=random(20); PridejKonec(x,df);
      until x=19;

(* tisk fronty + zruseni *)

      while df.celo<>df.konec do
      begin pom:=df.celo; write(df.celo^.klic:4); df.celo:=df.celo^.nasl;
            dispose(pom);
      end;
      nuluj(df.celo);
      dispose(df.celo);
      writeln;
      writeln(' znovu [A/N] ? '); z:=UpCase(readkey);
  until z ='N';
end.

(*  jak by se predchozi prepsalo v vyuzitim procedury OdeberCelo ? takto ? *)
(*    repeat OdeberCelo(x,df); write(x); until PrazdnaF(df);               *)
(*  je to spravne ?                                                        *)
(*  JE TU JEDNA CHYBA - jaka ?                                             *)
(*  nulovy prvek ? 							   *)
(*  vzdy zalezi na konkretnim napsani jednotlivych "nastroju"              *)
(*  a jejich provazanosti                                                  *)
